from __future__ import annotations

from fastapi_voyager.type_helper import (
    update_forward_refs,
    full_class_name,
    get_core_types,
    get_type_name
)
from fastapi_voyager.type import (
    FieldInfo,
    PK,
    FieldType,
    LinkType,
    Link,
    ModuleNode,
    SchemaNode,
)
from pydantic import BaseModel
from pydantic_resolve import ErDiagram, Entity, Relationship, MultipleRelationship
from logging import getLogger
from fastapi_voyager.module import build_module_schema_tree

logger = getLogger(__name__)


class DiagramRenderer:
    def __init__(
        self,
        *,
        show_fields: FieldType = 'single',
        show_module: bool = True
    ) -> None:
        self.show_fields = show_fields if show_fields in ('single', 'object', 'all') else 'single'
        self.show_module = show_module

        logger.info(f'show_module: {self.show_module}')

    def render_schema_label(self, node: SchemaNode, color: str | None=None) -> str:
        has_base_fields = any(f.from_base for f in node.fields)
        fields = [n for n in node.fields if n.from_base is False]

        if self.show_fields == 'all':
            _fields = fields
        elif self.show_fields == 'object':
            _fields = [f for f in fields if f.is_object is True]
        else:  # 'single'
            _fields = []

        fields_parts: list[str] = []
        if self.show_fields == 'all' and has_base_fields:
            fields_parts.append('<tr><td align="left" cellpadding="8"><font color="#999">  Inherited Fields ... </font></td></tr>')

        for field in _fields:
            type_name = field.type_name[:25] + '..' if len(field.type_name) > 25 else field.type_name
            display_xml = f'<s align="left">{field.name}: {type_name}</s>' if field.is_exclude else f'{field.name}: {type_name}'
            field_str = f"""<tr><td align="left" port="f{field.name}" cellpadding="8"><font>  {display_xml}    </font></td></tr>"""
            fields_parts.append(field_str)

        header_color = '#009485' if color is None else color
        header = f"""<tr><td cellpadding="6" bgcolor="{header_color}" align="center" colspan="1" port="{PK}"> <font color="white">    {node.name}    </font></td> </tr>"""
        field_content = ''.join(fields_parts) if fields_parts else ''
        return f"""<<table border="0" cellborder="1" cellpadding="0" cellspacing="0" bgcolor="white"> {header} {field_content}   </table>>"""

    def _handle_schema_anchor(self, source: str) -> str:
        if '::' in source:
            a, b = source.split('::', 1)
            return f'"{a}":{b}'
        return f'"{source}"'

    def render_link(self, link: Link) -> str:
        h = self._handle_schema_anchor
        if link.type == 'schema':
            return f"""{h(link.source)}:e -> {h(link.target)}:w [style = "solid", label = "{link.label}", minlen=3];"""
        else:
            raise ValueError(f'Unknown link type: {link.type}')

    def render_module_schema_content(self, nodes: list[SchemaNode]) -> str:
        def render_node(node: SchemaNode, color: str | None=None) -> str:
            return f'''
                "{node.id}" [
                    label = {self.render_schema_label(node, color)}
                    shape = "plain"
                    margin="0.5,0.1"
                ];'''

        def render_module_schema(mod: ModuleNode, show_cluster:bool=True) -> str:
            inner_nodes = [ render_node(node) for node in mod.schema_nodes ]
            inner_nodes_str = '\n'.join(inner_nodes)
            child_str = '\n'.join(render_module_schema(mod=m, show_cluster=show_cluster) for m in mod.modules)

            if show_cluster:
                return f'''
                    subgraph cluster_module_{mod.fullname.replace('.', '_')} {{
                        tooltip="{mod.fullname}"
                        color = "#666"
                        style="rounded"
                        label = "  {mod.name}"
                        labeljust = "l"
                        pencolor="#ccc"
                        penwidth=""
                        {inner_nodes_str}
                        {child_str}
                    }}'''
            else:
                return f'''
                    {inner_nodes_str}
                    {child_str}
                '''

        # if self.show_module:
        module_schemas = build_module_schema_tree(nodes)
        return '\n'.join(render_module_schema(mod=m, show_cluster=self.show_module) for m in module_schemas)

    def render_dot(self, nodes: list[SchemaNode], links: list[Link], spline_line=False) -> str:
        module_schemas_str = self.render_module_schema_content(nodes)
        link_str = '\n'.join(self.render_link(link) for link in links)

        dot_str = f'''
        digraph world {{
            pad="0.5"
            nodesep=0.8
            {'splines=line' if spline_line else ''}
            fontname="Helvetica,Arial,sans-serif"
            node [fontname="Helvetica,Arial,sans-serif"]
            edge [
                fontname="Helvetica,Arial,sans-serif"
                color="gray"
            ]
            graph [
                rankdir = "LR"
            ];
            node [
                fontsize = "16"
            ];

            subgraph cluster_schema {{
                color = "#aaa"
                margin=18
                style="dashed"
                label="  ER Diagram"
                labeljust="l"
                fontsize="20"
                    {module_schemas_str}
            }}

            {link_str}
            }}
        '''
        return dot_str


class VoyagerErDiagram:
    def __init__(self, 
                 er_diagram: ErDiagram, 
                 show_fields: FieldType = 'single',
                 show_module: bool = False):
        self.er_diagram = er_diagram
        self.nodes: list[SchemaNode] = []
        self.node_set: dict[str, SchemaNode] = {}

        self.links: list[Link] = []
        self.link_set: set[tuple[str, str]] = set()

        self.fk_set: dict[str, set[str]] = {}

        self.show_field = show_fields
        self.show_module = show_module
    
    def generate_node_head(self, link_name: str):
        return f'{link_name}::{PK}'

    def analysis_entity(self, entity: Entity):
        schema = entity.kls
        update_forward_refs(schema)
        self.add_to_node_set(schema, fk_set=self.fk_set.get(full_class_name(schema)))

        for relationship in entity.relationships:
            annos = get_core_types(relationship.target_kls)
            for anno in annos:
                self.add_to_node_set(anno, fk_set=self.fk_set.get(full_class_name(anno)))
                source_name = f'{full_class_name(schema)}::f{relationship.field}'
                if isinstance(relationship, Relationship):
                    self.add_to_link_set(
                        source=source_name,
                        source_origin=full_class_name(schema),
                        target=self.generate_node_head(full_class_name(anno)),
                        target_origin=full_class_name(anno),
                        type='schema',
                        label=get_type_name(relationship.target_kls))

                elif isinstance(relationship, MultipleRelationship):
                    for link in relationship.links:
                        self.add_to_link_set(
                            source=source_name,
                            source_origin=full_class_name(schema),
                            target=self.generate_node_head(full_class_name(anno)),
                            target_origin=full_class_name(anno),
                            type='schema',
                            biz=link.biz,
                            label=f'{get_type_name(relationship.target_kls)} / {link.biz} '
                        )

    def add_to_node_set(self, schema, fk_set: set[str] | None = None) -> str:
        """
        1. calc full_path, add to node_set
        2. if duplicated, do nothing, else insert
        2. return the full_path
        """
        full_name = full_class_name(schema)

        if full_name not in self.node_set:
            # skip meta info for normal queries
            self.node_set[full_name] = SchemaNode(
                id=full_name, 
                module=schema.__module__,
                name=schema.__name__,
                fields=get_fields(schema, fk_set)
            )
        return full_name

    def add_to_link_set(
            self, 
            source: str, 
            source_origin: str,
            target: str, 
            target_origin: str,
            type: LinkType,
            label: str,
            biz: str | None = None
        ) -> bool:
        """
        1. add link to link_set
        2. if duplicated, do nothing, else insert
        """
        pair = (source, target, biz)
        if result := pair not in self.link_set:
            self.link_set.add(pair)
            self.links.append(Link(
                source=source,
                source_origin=source_origin,
                target=target,
                target_origin=target_origin,
                type=type,
                label=label
            ))
        return result


    def render_dot(self):
        self.fk_set = {
            full_class_name(entity.kls): set([rel.field for rel in entity.relationships])
                for entity in self.er_diagram.configs
        }

        for entity in self.er_diagram.configs:
            self.analysis_entity(entity)
        renderer = DiagramRenderer(show_fields=self.show_field, show_module=self.show_module)
        return renderer.render_dot(list(self.node_set.values()), self.links)


def get_fields(schema: type[BaseModel], fk_set: set[str] | None = None) -> list[FieldInfo]:

    fields: list[FieldInfo] = []
    for k, v in schema.model_fields.items():
        anno = v.annotation
        fields.append(FieldInfo(
            is_object=k in fk_set if fk_set is not None else False,
            name=k,
            from_base=False,
            type_name=get_type_name(anno),
            is_exclude=bool(v.exclude)
        ))
    return fields



from fastapi_voyager.filter import filter_subgraph_by_module_prefix
from fastapi_voyager.type import PK, Link, Route, SchemaNode, Tag


def _make_tag_route_link(tag: Tag, route: Route) -> Link:
    return Link(
        source=tag.id,
        source_origin=tag.id,
        target=route.id,
        target_origin=route.id,
        type="tag_route",
    )


def test_filter_subgraph_filters_nodes_and_links():
    tag = Tag(id="tag1", name="Tag 1", routes=[])
    route = Route(id="route1", name="route1", module="api.routes")
    tag.routes.append(route)

    node_a = SchemaNode(id="pkg.ModelA", name="ModelA", module="pkg.moduleA")
    node_b = SchemaNode(id="pkg.ModelB", name="ModelB", module="target.moduleB")

    links = [
        _make_tag_route_link(tag, route),
        Link(
            source=route.id,
            source_origin=route.id,
            target=f"{node_a.id}::{PK}",
            target_origin=node_a.id,
            type="route_to_schema",
        ),
        Link(
            source=f"{node_a.id}::ffield",
            source_origin=node_a.id,
            target=f"{node_b.id}::{PK}",
            target_origin=node_b.id,
            type="schema",
        ),
    ]

    tags = [tag]
    routes = [route]
    nodes = [node_a, node_b]

    _, _, filtered_nodes, filtered_links = filter_subgraph_by_module_prefix(
        tags=tags,
        routes=routes,
        links=links,
        nodes=nodes,
        module_prefix="target",
    )

    assert filtered_nodes == [node_b]
    assert any(
        lk.type == "route_to_schema" and \
        lk.source_origin == route.id and \
        lk.target_origin == node_b.id
        for lk in filtered_links
    )
    assert len(filtered_links) == 2  # tag -> route and merged route -> filtered node



def test_filter_subgraph_handles_cycles_and_multiple_matches():
    tag = Tag(id="tag-main", name="Tag", routes=[])
    route = Route(id="route-main", name="route", module="api.routes")
    tag.routes.append(route)

    node_root = SchemaNode(id="pkg.Root", name="Root", module="pkg.root")
    node_mid = SchemaNode(id="pkg.Mid", name="Mid", module="pkg.mid")
    node_target1 = SchemaNode(id="pkg.Target1", name="Target1", module="target.mod.alpha")
    node_target2 = SchemaNode(id="pkg.Target2", name="Target2", module="target.mod.beta")

    links = [
        _make_tag_route_link(tag, route),
        Link(
            source=route.id,
            source_origin=route.id,
            target=f"{node_root.id}::{PK}",
            target_origin=node_root.id,
            type="route_to_schema",
        ),
        Link(
            source=f"{node_root.id}::ffield",
            source_origin=node_root.id,
            target=f"{node_mid.id}::{PK}",
            target_origin=node_mid.id,
            type="schema",
        ),
        Link(
            source=f"{node_mid.id}::{PK}",
            source_origin=node_mid.id,
            target=f"{node_target1.id}::{PK}",
            target_origin=node_target1.id,
            type="parent",
        ),
        Link(
            source=f"{node_mid.id}::ffield",
            source_origin=node_mid.id,
            target=f"{node_target2.id}::{PK}",
            target_origin=node_target2.id,
            type="subset",
        ),
        Link(
            source=f"{node_target1.id}::ffield",
            source_origin=node_target1.id,
            target=f"{node_root.id}::{PK}",
            target_origin=node_root.id,
            type="schema",
        ),
    ]

    nodes = [node_root, node_mid, node_target1, node_target2]

    _, _, filtered_nodes, filtered_links = filter_subgraph_by_module_prefix(
        tags=[tag],
        routes=[route],
        links=links,
        nodes=nodes,
        module_prefix="target.mod",
    )

    assert filtered_nodes == [node_target1, node_target2]

    route_to_schema_targets = {
        (lk.source_origin, lk.target_origin)
        for lk in filtered_links
        if lk.type == "route_to_schema"
    }
    assert route_to_schema_targets == {
        (route.id, node_target1.id),
        (route.id, node_target2.id),
    }

    assert all(lk.type in {"tag_route", "route_to_schema"} for lk in filtered_links)
    assert len(filtered_links) == 3  # 1 tag_route + 2 merged links



from __future__ import annotations
from typing import Literal
from pydantic import BaseModel
from pydantic_resolve import Relationship, MultipleRelationship, Link
from .base_entity import BaseEntity



class Member(BaseModel):
    id: int
    first_name: str
    last_name: str

class Task(BaseModel, BaseEntity):
    __pydantic_resolve_relationships__ = [
        Relationship(field='owner_id', target_kls=Member),
        Relationship(field='story_id', target_kls='Story'),
    ]
    id: int
    story_id: int
    description: str
    owner_id: int

class Story(BaseModel, BaseEntity):
    __pydantic_resolve_relationships__ = [
        Relationship(field='id', target_kls=list[Task]),
    ]
    id: int
    type: Literal['feature', 'bugfix']
    dct: dict
    sprint_id: int
    title: str
    description: str

class Sprint(BaseModel, BaseEntity):
    __pydantic_resolve_relationships__ = [
        MultipleRelationship(field='id', target_kls=list[Story], links=[
            Link(biz='all'),
            Link(biz='done'),
        ]) 
    ]
    id: int
    name: str



import pydantic_resolve.constant as const
from fastapi import FastAPI, routing
from pydantic import BaseModel

from fastapi_voyager.filter import (
    filter_graph,
    filter_subgraph_by_module_prefix,
    filter_subgraph_from_tag_to_schema_by_module_prefix,
)
from fastapi_voyager.render import Renderer
from fastapi_voyager.type import PK, CoreData, FieldType, Link, LinkType, Route, SchemaNode, Tag
from fastapi_voyager.type_helper import (
    full_class_name,
    get_bases_fields,
    get_core_types,
    get_pydantic_fields,
    get_type_name,
    is_inheritance_of_pydantic_base,
    is_non_pydantic_type,
    update_forward_refs,
)


class Voyager:
    def __init__(
            self, 
            schema: str | None = None, 
            schema_field: str | None = None,
            show_fields: FieldType = 'single',
            include_tags: list[str] | None = None,
            module_color: dict[str, str] | None = None,
            route_name: str | None = None,
            hide_primitive_route: bool = False,
            show_module: bool = True,
        ):

        self.routes: list[Route] = []

        self.nodes: list[SchemaNode] = []
        self.node_set: dict[str, SchemaNode] = {}

        self.link_set: set[tuple[str, str]] = set()
        self.links: list[Link] = []

        # store Tag by id, and also keep a list for rendering order
        self.tag_set: dict[str, Tag] = {}
        self.tags: list[Tag] = []

        self.include_tags = include_tags
        self.schema = schema
        self.schema_field = schema_field
        self.show_fields = show_fields if show_fields in ('single','object','all') else 'object'
        self.module_color = module_color or {}
        self.route_name = route_name
        self.hide_primitive_route = hide_primitive_route
        self.show_module = show_module
    

    def _get_available_route(self, app: FastAPI):
        for route in app.routes:
            if isinstance(route, routing.APIRoute):
                yield route


    def analysis(self, app: FastAPI):
        """
        1. get routes which return pydantic schema
            1.1 collect tags and routes, add links tag-> route
            1.2 collect response_model and links route -> response_model

        2. iterate schemas, construct the schema/model nodes and their links
        """
        schemas: list[type[BaseModel]] = []

        # First, group all routes by tag
        routes_by_tag: dict[str, list] = {}
        for route in self._get_available_route(app):
            tags = getattr(route, 'tags', None)

            # using multiple tags is harmful, it's not recommended and will not be supported
            route_tag = tags[0] if tags else '__default__'  
            routes_by_tag.setdefault(route_tag, []).append(route)

        # Then filter by include_tags if provided
        if self.include_tags:
            filtered_routes_by_tag = {tag: routes for tag, routes in routes_by_tag.items() 
                                    if tag in self.include_tags}
        else:
            filtered_routes_by_tag = routes_by_tag

        # Process filtered routes
        for route_tag, routes in filtered_routes_by_tag.items():

            tag_id = f'tag__{route_tag}'
            tag_obj = Tag(id=tag_id, name=route_tag, routes=[])
            self.tags.append(tag_obj)

            for route in routes:
                # add route and create links
                route_id = full_class_name(route.endpoint)
                route_name = route.endpoint.__name__
                route_module = route.endpoint.__module__

                # filter by route_name (route.id) if provided
                if self.route_name is not None and route_id != self.route_name:
                    continue

                is_primitive_response = is_non_pydantic_type(route.response_model)
                # filter primitive route if needed
                if self.hide_primitive_route and is_primitive_response:
                    continue

                self.links.append(Link(
                    source=tag_id,
                    source_origin=tag_id,
                    target=route_id,
                    target_origin=route_id,
                    type='tag_route'
                ))

                route_obj = Route(
                    id=route_id,
                    name=route_name,
                    module=route_module,
                    unique_id=route.operation_id or route.unique_id,
                    response_schema=get_type_name(route.response_model),
                    is_primitive=is_primitive_response
                )
                self.routes.append(route_obj)
                tag_obj.routes.append(route_obj)

                # add response_models and create links from route -> response_model
                for schema in get_core_types(route.response_model):
                    if schema and issubclass(schema, BaseModel):
                        is_primitive_response = False
                        target_name = full_class_name(schema)
                        self.links.append(Link(
                            source=route_id,
                            source_origin=route_id,
                            target=self.generate_node_head(target_name),
                            target_origin=target_name,
                            type='route_to_schema'
                        ))

                        schemas.append(schema)

        for s in schemas:
            self.analysis_schemas(s)
        
        self.nodes = list(self.node_set.values())


    def add_to_node_set(self, schema):
        """
        1. calc full_path, add to node_set
        2. if duplicated, do nothing, else insert
        2. return the full_path
        """
        full_name = full_class_name(schema)
        bases_fields = get_bases_fields([s for s in schema.__bases__ if is_inheritance_of_pydantic_base(s)])
        
        subset_reference = getattr(schema, const.ENSURE_SUBSET_REFERENCE, None)
        if subset_reference and is_inheritance_of_pydantic_base(subset_reference):
            bases_fields.update(get_bases_fields([subset_reference]))

        if full_name not in self.node_set:
            # skip meta info for normal queries
            self.node_set[full_name] = SchemaNode(
                id=full_name, 
                module=schema.__module__,
                name=schema.__name__,
                fields=get_pydantic_fields(schema, bases_fields)
            )
        return full_name


    def add_to_link_set(
            self, 
            source: str, 
            source_origin: str,
            target: str, 
            target_origin: str,
            type: LinkType
        ) -> bool:
        """
        1. add link to link_set
        2. if duplicated, do nothing, else insert
        """
        pair = (source, target)
        if result := pair not in self.link_set:
            self.link_set.add(pair)
            self.links.append(Link(
                source=source,
                source_origin=source_origin,
                target=target,
                target_origin=target_origin,
                type=type
            ))
        return result


    def analysis_schemas(self, schema: type[BaseModel]):
        """
        1. cls is the source, add schema
        2. pydantic fields are targets, if annotation is subclass of BaseMode, add fields and add links
        3. recursively run walk_schema
        """
        
        update_forward_refs(schema)
        self.add_to_node_set(schema)

        base_fields = set()

        # handle schema inside ensure_subset(schema)
        if subset_reference := getattr(schema,  const.ENSURE_SUBSET_REFERENCE, None):
            if is_inheritance_of_pydantic_base(subset_reference):

                self.add_to_node_set(subset_reference)
                self.add_to_link_set(
                    source=self.generate_node_head(full_class_name(schema)),
                    source_origin=full_class_name(schema),
                    target= self.generate_node_head(full_class_name(subset_reference)), 
                    target_origin=full_class_name(subset_reference),
                    type='subset')
                self.analysis_schemas(subset_reference)

        # handle bases
        for base_class in schema.__bases__:
            if is_inheritance_of_pydantic_base(base_class):
                # collect base class field names to avoid duplicating inherited fields
                try:
                    base_fields.update(getattr(base_class, 'model_fields', {}).keys())
                except Exception:
                    # be defensive in case of unconventional BaseModel subclasses
                    pass
                self.add_to_node_set(base_class)
                self.add_to_link_set(
                    source=self.generate_node_head(full_class_name(schema)),
                    source_origin=full_class_name(schema),
                    target=self.generate_node_head(full_class_name(base_class)),
                    target_origin=full_class_name(base_class),
                    type='parent')
                self.analysis_schemas(base_class)

        # handle fields
        for k, v in schema.model_fields.items():
            # skip fields inherited from base classes
            if k in base_fields:
                continue
            annos = get_core_types(v.annotation)
            for anno in annos:
                if anno and is_inheritance_of_pydantic_base(anno):
                    self.add_to_node_set(anno)
                    # add f prefix to fix highlight issue in vsc graphviz interactive previewer
                    source_name = f'{full_class_name(schema)}::f{k}'
                    if self.add_to_link_set(
                        source=source_name,
                        source_origin=full_class_name(schema),
                        target=self.generate_node_head(full_class_name(anno)),
                        target_origin=full_class_name(anno),
                        type='schema'):
                        self.analysis_schemas(anno)


    def generate_node_head(self, link_name: str):
        return f'{link_name}::{PK}'

    def dump_core_data(self):
        _tags, _routes, _nodes, _links = filter_graph(
            schema=self.schema,
            schema_field=self.schema_field,
            tags=self.tags,
            routes=self.routes,
            nodes=self.nodes,
            links=self.links,
            node_set=self.node_set,
        )
        return CoreData(
            tags=_tags,
            routes=_routes,
            nodes=_nodes,
            links=_links,
            show_fields=self.show_fields,
            module_color=self.module_color,
            schema=self.schema
        )

    def handle_hide(self, tags, routes, links):
        if self.include_tags:
            return [], routes, [lk for lk in links if lk.type != 'tag_route']
        else:
            return tags, routes, links
    
    def calculate_filtered_tag_and_route(self):
        _tags, _routes, _, _ = filter_graph(
            schema=self.schema,
            schema_field=self.schema_field,
            tags=self.tags,
            routes=self.routes,
            nodes=self.nodes,
            links=self.links,
            node_set=self.node_set,
        )
        # filter tag.routes based by _routes
        route_ids = {r.id for r in _routes}
        for t in _tags:
            t.routes = [r for r in t.routes if r.id in route_ids]
        return _tags

    def render_dot(self):
        _tags, _routes, _nodes, _links = filter_graph(
            schema=self.schema,
            schema_field=self.schema_field,
            tags=self.tags,
            routes=self.routes,
            nodes=self.nodes,
            links=self.links,
            node_set=self.node_set,
        )

        renderer = Renderer(show_fields=self.show_fields, module_color=self.module_color, schema=self.schema, show_module=self.show_module)

        _tags, _routes, _links = self.handle_hide(_tags, _routes, _links)
        return renderer.render_dot(_tags, _routes, _nodes, _links)
    

    def render_tag_level_brief_dot(self, module_prefix: str | None = None):
        _tags, _routes, _nodes, _links = filter_graph(
            schema=self.schema,
            schema_field=self.schema_field,
            tags=self.tags,
            routes=self.routes,
            nodes=self.nodes,
            links=self.links,
            node_set=self.node_set,
        )

        _tags, _routes, _nodes, _links = filter_subgraph_by_module_prefix(
            module_prefix=module_prefix,
            tags=_tags,
            routes=_routes,
            nodes=_nodes,
            links=_links,
        )

        renderer = Renderer(show_fields=self.show_fields, module_color=self.module_color, schema=self.schema, show_module=self.show_module)

        _tags, _routes, _links = self.handle_hide(_tags, _routes, _links)
        return renderer.render_dot(_tags, _routes, _nodes, _links, True)

    def render_overall_brief_dot(self, module_prefix: str | None = None):
        _tags, _routes, _nodes, _links = filter_graph(
            schema=self.schema,
            schema_field=self.schema_field,
            tags=self.tags,
            routes=self.routes,
            nodes=self.nodes,
            links=self.links,
            node_set=self.node_set,
        )

        _tags, _routes, _nodes, _links = filter_subgraph_from_tag_to_schema_by_module_prefix(
            module_prefix=module_prefix,
            tags=_tags,
            routes=_routes,
            nodes=_nodes,
            links=_links,
        )

        renderer = Renderer(show_fields=self.show_fields, module_color=self.module_color, schema=self.schema, show_module=self.show_module)

        _tags, _routes, _links = self.handle_hide(_tags, _routes, _links)
        return renderer.render_dot(_tags, _routes, _nodes, _links, True)